@@ -2,6 +2,9 @@ use std::collections::VecDeque;
 use std::iter;
 use std::f64;
 use std::rc::Rc;
+use std::sync::mpsc::{self, channel};
+use rayon::prelude::*;
+use rayon;
 use polytype::{Context, Type};
 
 use super::{Expression, Language, LinkedList};
@@ -10,26 +13,72 @@ const BUDGET_INCREMENT: f64 = 1.0;
 const MAX_DEPTH: u32 = 256;
 
 pub fn new<'a>(dsl: &'a Language, request: Type) -> Box<Iterator<Item = (Expression, f64)> + 'a> {
-    let budget = |offset: f64| (offset, offset + BUDGET_INCREMENT);
+    let to_budget = |offset: f64| (offset, offset + BUDGET_INCREMENT);
     let ctx = Context::default();
     let env = Rc::new(LinkedList::default());
-    let depth = 0;
     Box::new(
         (0..)
             .map(|n| BUDGET_INCREMENT * f64::from(n))
-            .flat_map(move |offset| {
-                enumerate(
-                    dsl,
-                    request.clone(),
-                    &ctx,
-                    env.clone(),
-                    budget(offset),
-                    depth,
-                ).map(move |(log_prior, _, expr)| (expr, log_prior))
-            }),
+            .map(to_budget)
+            .zip(iter::repeat(request))
+            .flat_map(move |(budget, request)| {
+                enumerate(dsl, request, &ctx, env.clone(), budget, 0)
+            })
+            .map(|(log_prior, _, expr)| (expr, log_prior)),
     )
 }
 
+/// enumerate expressions in parallel within the budget interval
+#[allow(dead_code)]
+fn new_par(dsl: Language, request: Type, budget: (f64, f64)) -> mpsc::IntoIter<(Expression, f64)> {
+    let (tx, rx) = channel();
+    rayon::spawn(move || {
+        rayon::iter::repeat(tx)
+            .zip(exponential_decay(budget))
+            .for_each(|(tx, budget)| {
+                let ctx = Context::default();
+                let env = Rc::new(LinkedList::default());
+                let e = enumerate(&dsl, request.clone(), &ctx, env.clone(), budget, 0)
+                    .map(|(log_prior, _, expr)| (expr, log_prior));
+                for (expr, logprior) in e {
+                    if tx.send((expr, logprior)).is_err() {
+                        // receiving end was dropped
+                        break;
+                    }
+                }
+            })
+    });
+    rx.into_iter()
+}
+
+#[allow(dead_code)]
+fn exponential_decay(budget: (f64, f64)) -> Vec<(f64, f64)> {
+    // because depth values correspond to description length in nats, we
+    // assume that for pieces to have the same total description coverage
+    // (which should correspond roughly to enumerate time), their budget
+    // widths follow exponential decay.
+    let pieces = match budget.1 {
+        x if x < 4. => 2,
+        x if x < 8. => 4,
+        x if x < 10. => 8,
+        x if x < 12. => 16,
+        x if x < 15. => 32,
+        x if x < 17. => 64,
+        _ => 256,
+    };
+    let step = (budget.1.exp() - budget.0.exp()) / (f64::from(pieces));
+    let mut v = Vec::new();
+    let mut prev = budget.0;
+    for x in (1..(pieces + 1))
+        .map(|i| budget.0.exp() + (f64::from(i)) * step)
+        .map(f64::ln)
+    {
+        v.push((prev, x));
+        prev = x;
+    }
+    v
+}
+
 pub fn likelihood<'a>(dsl: &'a Language, request: &Type, expr: &Expression) -> f64 {
     let ctx = Context::default();
     let env = Rc::new(LinkedList::default());
@@ -43,9 +92,9 @@ fn likelihood_internal<'a>(
     mut expr: &Expression,
 ) -> (f64, Context) {
     if let Type::Arrow(ref arrow) = *request {
-        let env = LinkedList::prepend(env, *arrow.arg.clone());
+        let env = LinkedList::prepend(env, arrow.arg.clone());
         if let Expression::Abstraction(ref body) = *expr {
-            likelihood_internal(dsl, &*arrow.ret, ctx, &env, body)
+            likelihood_internal(dsl, &arrow.ret, ctx, &env, body)
         } else {
             (f64::NEG_INFINITY, ctx.clone()) // invalid expression
         }
@@ -98,8 +147,8 @@ fn enumerate<'a>(
     if budget.1 <= 0f64 || depth > MAX_DEPTH {
         Box::new(iter::empty())
     } else if let Type::Arrow(arrow) = request {
-        let env = LinkedList::prepend(&env, *arrow.arg);
-        let it = enumerate(dsl, *arrow.ret, ctx, env, budget, depth)
+        let env = LinkedList::prepend(&env, arrow.arg.clone());
+        let it = enumerate(dsl, arrow.ret, ctx, env, budget, depth)
             .map(|(ll, ctx, body)| (ll, ctx, Expression::Abstraction(Box::new(body))));
         Box::new(it)
     } else {
